package edu.bu.signstream.common.db.ss4;

import edu.bu.signstream.common.graph.Graph;
import edu.bu.signstream.common.util.Pair;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.Panel;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3CodingScheme;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.ViewMode;
import edu.bu.signstream.common.util.vo.ss3.db.SS3Database;
import edu.bu.signstream.grepresentation.view.ViewData;
import edu.bu.signstream.grepresentation.view.ViewDataManager;
import edu.bu.signstream.media.Audio;
import edu.bu.signstream.media.AudioInfo;
import edu.bu.signstream.media.Video;
import edu.bu.signstream.media.VideoInfo;
import edu.bu.signstream.time.*;

import org.w3c.dom.CharacterData;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import java.net.URI;
import java.util.*;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

public class SS4Util {
    // Prefixing attribute and tag names
    public static String prefixAttr(String attr, String pre) { return !"".equals(pre) ? pre + "_" + attr : attr; }
    public static String prefixTag(String tag, String pre) { return !"".equals(pre) ? pre + "-" + tag : tag; }



    // Checking if elements have attributes
    public static boolean has(Element el, String attr) { return el.hasAttribute(attr); }
    public static boolean hasRef(Element el, String prefix, String tag) {
        return has(el, prefixAttr("ID", prefixAttr(tag, prefix)));
    }
    public static boolean hasRef(Element el, String tag) { return hasRef(el, "", tag); }

    // Writing primitives
    public static Element write(Element el, String attr, String value) {
        el.setAttribute(attr, value);
        return el;
    }
    public static Element writeLong(Element el, String attr, long value) {
        return write(el, attr, Long.toString(value));
    }
    public static Element writeDouble(Element el, String attr, double value) {
        return write(el, attr, Double.toString(value));
    }

    // Reading primitives
    public static Optional<String> readOpt(Element el, String attr) {
        return Optional.ofNullable(el.getAttribute(attr));
    }
    public static String read(Element el, String attr) { return readOpt(el, attr).orElseThrow(); }
    public static Optional<Long> readLongOpt(Element el, String attr) {
        return readOpt(el, attr)
            .flatMap(s -> {
                try {
                    return Optional.of(Long.parseLong(s));
                } catch (NumberFormatException e) {
                    return Optional.empty();
                }
            });
    }
    public static long readLong(Element el, String attr) { return readLongOpt(el, attr).orElseThrow(); }
    public static Optional<Double> readDoubleOpt(Element el, String attr) {
        return readOpt(el, attr)
            .flatMap(s -> {
                try {
                    return Optional.of(Double.parseDouble(s));
                } catch (NumberFormatException e) {
                    return Optional.empty();
                }
            });
    }
    public static double readDouble(Element el, String attr) { return readDoubleOpt(el, attr).orElseThrow(); }
    public static Optional<Integer> readIntegerOpt(Element el, String attr) {
        return readOpt(el, attr)
                .flatMap(s -> {
                    try {
                        return Optional.of(Integer.parseInt(s));
                    } catch (NumberFormatException e) {
                        return Optional.empty();
                    }
                });
        }
        public static double readInteger(Element el, String attr) { return readIntegerOpt(el, attr).orElseThrow(); }


    public static Optional<Element> readElOpt(Document doc, String tag) {
        var nodes = doc.getElementsByTagName("SIGNSTREAM-DATABASE");
        return Optional.ofNullable((Element)nodes.item(0));
    }


    // Writing child elements
    // Reading child elements
    public static Optional<Element> readChildOpt(Element el, String tag) {
        var descendants = el.getElementsByTagName(tag);

        // the filter ensures only direct descendants (immediate children of Element el)
        var children = IntStream.range(0, descendants.getLength())
                .filter(i -> descendants.item(i).getParentNode() != null && descendants.item(i).getParentNode().equals(el))
                .mapToObj(i -> (Element) descendants.item(i))
                .collect(Collectors.toCollection(ArrayList::new));
        
        return children.size() > 0 ? Optional.ofNullable((Element) children.get(0)) : Optional.empty();
    }
    public static Element readChild(Element el, String tag) { return readChildOpt(el, tag).orElseThrow(); }

    public static Stream<Element> readChildren(Element el, String tag) {
        var descendants = el.getElementsByTagName(tag);

        // the filter ensures only direct descendants (immediate children of Element el)
        return IntStream.range(0, descendants.getLength())
            .filter(i -> descendants.item(i).getParentNode() != null && descendants.item(i).getParentNode().equals(el))
            .mapToObj(i -> (Element) descendants.item(i));
    }
    public static Stream<Element> readChildrenOrdered(Element el, String tag) {
        return readChildren(el, tag)
            .sorted(Comparator.comparing(e -> readLongOpt(e, "ORDER").orElse(Long.MAX_VALUE)));
    }

    // Writing nested children in a list
    public static Element writeChildList(Element el, String tag, Stream<Element> children) {
        var container = el.getOwnerDocument().createElement(tag);
        children.forEach(container::appendChild);
        el.appendChild(container);
        return el;
    }
    public static Element writeChildListOrdered(Element el, String tag, Stream<Element> children) {
        var container = el.getOwnerDocument().createElement(tag);

        var counter = new AtomicLong(1);
        children.forEach(c -> {
            writeLong(c, "ORDER", counter.getAndIncrement());
            container.appendChild(c);
        });

        el.appendChild(container);
        return el;
    }

    // Reading nested children in a list
    public static Optional<Stream<Element>> readChildListOpt(Element el, String tag, String childTag) {
        var container = readChildOpt(el, tag);
        return container.map(c -> readChildren(c, childTag));
    }
    public static Optional<Stream<Element>> readChildListOrderedOpt(Element el, String tag, String childTag) {
        var container = readChildOpt(el, tag);
        return container.map(c -> readChildrenOrdered(c, childTag));
    }
    public static Stream<Element> readChildList(Element el, String tag, String childTag) {
        return readChildListOpt(el, tag, childTag).orElseThrow();
    }
    public static Stream<Element> readChildListOrdered(Element el, String tag, String childTag) {
        return readChildListOrderedOpt(el, tag, childTag).orElseThrow();
    }




    // Writing One-to-One references
    public static Element writeRef(Element el, String prefix, String tag, String id) {
        return write(el, prefixAttr("ID", prefixAttr(tag, prefix)), id);
    }
    public static Element writeRef(Element el, String tag, String id) { return writeRef(el, "", tag, id); }

    // Reading One-to-One References
    public static Optional<String> readRefOpt(Element el, String prefix, String tag) {
        return readOpt(el, prefixAttr("ID", prefixAttr(tag, prefix)));
    }
    public static Optional<String> readRefOpt(Element el, String tag) { return readRefOpt(el, "", tag); }
    public static String readRef(Element el, String prefix, String tag) { return readRefOpt(el, prefix, tag).orElseThrow(); }
    public static String readRef(Element el, String tag) {
        return readRefOpt(el, tag).orElseThrow();
    }

    public static Element writeRefs(Element el, String prefix, String tag, Stream<String> ids) {
        var doc = el.getOwnerDocument();
        var container = doc.createElement(prefixTag(prefixTag("REFS", tag), prefix));

        ids.forEach(id -> {
            var ref = doc.createElement("REF");
            writeRef(ref, tag, id);
            container.appendChild(ref);
        });
        el.appendChild(container);
        return el;
    }

    /**
     * @param el The element to write to
     * @param prefix The prefix of the XML, in this case, VISIBLE
     * @param tag The tag to use for the XML, in this case, PANEL... Together they make VISIBLE-PANEL and REFS, a string constant in the method itself. I don't know why it is done this way.
     * @param viewDataManager The ViewDataManager object that contains the list of ViewData objects
     * @return The element with the ViewDataManager object written to it
     */
    // Writing One-to-Many references
    public static Element writeRefs(Element el, String prefix, String tag, ViewDataManager viewDataManager) {
        Document doc = el.getOwnerDocument();
        Element container = doc.createElement(prefixTag(prefixTag("REFS", tag), prefix)); // VISIBLE-PANEL-REFS is what the prefixTag method call ends up returning, it is very confusing

        if (viewDataManager != null) {
            for (ViewData viewData : viewDataManager.getViewDataList()) {
                Element ref = doc.createElement("REF"); // Create a new REF element that will be written to the XML
                write(ref, "PANEL_ID", viewData.getPanelID());
                write(ref, "PANEL_PERCENTAGE", String.valueOf(viewData.getPanelPercentage()));
                write(ref, "ORDER", String.valueOf(viewData.getOrder()));
                container.appendChild(ref);
            }

            el.appendChild(container);

            return el;
        }

        return null;
    }
    public static Element writeRefs(Element el, String tag, Stream<String> ids) { return writeRefs(el, "", tag, ids); }

    public static Element writeRefsOrdered(Element el, String prefix, String tag, Stream<String> ids) {
        var doc = el.getOwnerDocument();
        var container = doc.createElement(prefixTag(prefixTag("REFS", tag), prefix));

        var counter = new AtomicLong(1);
        ids.forEach(id -> {
            var ref = doc.createElement("REF");
            writeRef(ref, tag, id);
            writeLong(ref, "ORDER", counter.getAndIncrement());
            container.appendChild(ref);
        });
        return el;
    }
    public static Element writeRefsOrdered(Element el, String tag, Stream<String> ids) { return writeRefsOrdered(el, "", tag, ids); }

    // Reading One-to-Many References
    public static Optional<Stream<String>> readRefsOpt(Element el, String prefix, String tag) {
        var container = readChildOpt(el, prefixTag(prefixTag("REFS", tag), prefix)).orElse(null);
        if (container == null)
            return Optional.empty();

        return Optional.of(readChildren(container, "REF")
            .map(ref -> readRef(ref, tag)));
    }
    public static Optional<Stream<String>> readRefsOpt(Element el, String tag) { return readRefsOpt(el, "", tag); }
    public static Stream<String> readRefs(Element el, String prefix, String tag) { return readRefsOpt(el, prefix, tag).orElseThrow(); }
    public static Stream<String> readRefs(Element el, String tag) { return readRefs(el, "", tag); }

    public static Optional<Stream<String>> readRefsOrderedOpt(Element el, String prefix, String tag) {
        var container = readChildOpt(el, prefixTag(prefixTag("REFS", tag), prefix)).orElse(null);
        if (container == null)
            return Optional.empty();

        return Optional.of(readChildren(container, "REF")
            .map(ref -> {
                var id = readRef(ref, tag);
                var order = readLongOpt(el, "ORDER")
                    .orElse(Long.MAX_VALUE);
                return Pair.of(id, order);
            })
            .sorted(Comparator.comparing(p -> p.second))
            .map(p -> p.first));
    }
    public static Optional<Stream<String>> readRefsOrderedOpt(Element el, String tag) { return readRefsOrderedOpt(el, "", tag); }
    public static Stream<String> readRefsOrdered(Element el, String prefix, String tag) { return readRefsOrderedOpt(el, prefix, tag).orElseThrow(); }
    public static Stream<String> readRefsOrdered(Element el, String tag) { return readRefsOrdered(el, "", tag); }




    // Writing Times, Durations, and Intervals
    public static Element writeTime(Element el, String attr, Time t) {
        writeRef(el, attr, TIMERANGE_TAG, t.range().id());
        writeLong(el, attr, t.value());
        return el;
    }
    public static Element writeTime(Element el, Time t) { return writeTime(el, "TIME", t); }

    public static Element writeDuration(Element el, String attr, Duration d) {
        writeRef(el, attr, TIMEUNIT_TAG, d.unit().id());
        writeLong(el, attr, d.value());
        return el;
    }
    public static Element writeDuration(Element el, Duration d) { return writeDuration(el, "DURATION", d); }

    public static Element writeInterval(Element el, String attr, Interval i) {
        writeTime(el, prefixAttr("START", attr), i.t0);
        writeTime(el, prefixAttr("END", attr), i.t1);
        return el;
    }
    public static Element writeInterval(Element el, Interval i) { return writeInterval(el, "", i); }

    // Reading Times, Durations, and Intervals
    public static Optional<Time> readTimeOpt(SS3Database db, Element el, String attr) {
        var range = readRefOpt(el, attr, TIMERANGE_TAG)
            .flatMap(r -> Optional.ofNullable(db.getTimeRange(r)))
            .orElse(null);
        var value = readLongOpt(el, attr)
            .orElse(null);

        if (range == null || value == null)
            return Optional.empty();

        return Optional.of(range.time(value, range.unit()));
    }
    public static Optional<Time> readTimeOpt(SS3Database db, Element el) { return readTimeOpt(db, el, "TIME"); }
    public static Time readTime(SS3Database db, Element el, String attr) { return readTimeOpt(db, el, attr).orElseThrow(); }
    public static Time readTime(SS3Database db, Element el) { return readTimeOpt(db, el).orElseThrow(); }

    public static Optional<Duration> readDurationOpt(SS3Database db, Element el, String attr) {
        var unit = readRefOpt(el, attr, TIMEUNIT_TAG)
            .map(db::getTimeUnit)
            .orElse(null);
        var value = readLongOpt(el, attr)
            .orElse(null);

        if (unit == null || value == null)
            return Optional.empty();

        return Optional.of(new Duration(value, unit));
    }
    public static Optional<Duration> readDurationOpt(SS3Database db, Element el) { return readDurationOpt(db, el, "DURATION"); }
    public static Duration readDuration(SS3Database db, Element el, String attr) { return readDurationOpt(db, el, attr).orElseThrow(); }
    public static Duration readDuration(SS3Database db, Element el) { return readDurationOpt(db, el).orElseThrow(); }

    public static Optional<Interval> readIntervalOpt(SS3Database db, Element el, String attr) {
        var start = readTimeOpt(db, el, prefixAttr("START", attr))
            .orElse(null);
        var end = readTimeOpt(db, el, prefixAttr("END", attr))
            .orElse(null);

        if (start == null || end == null)
            return Optional.empty();

        return Optional.of(start.to(end));
    }
    public static Optional<Interval> readIntervalOpt(SS3Database db, Element el) { return readIntervalOpt(db, el, ""); }
    public static Interval readInterval(SS3Database db, Element el, String attr) { return readIntervalOpt(db, el, attr).orElseThrow(); }
    public static Interval readInterval(SS3Database db, Element el) { return readIntervalOpt(db, el).orElseThrow(); }




    // TODO Writing TimeRanges and Units
    public static final String TIMERANGE_LIST_TAG = "TIMERANGES";
    public static final String TIMERANGE_TAG = "TIMERANGE";
    public static Element createTimeRangeEl(Document doc, TimeRange range) {
        var el = doc.createElement(TIMERANGE_TAG);
        write(el, "ID", range.id());
        write(el, "NAME", range.name());
        if (range.parent() != null)
            write(el, "PARENT", range.parent().name());
        writeRef(el, TIMEUNIT_TAG, range.unit().id());
        writeDuration(el, "OFFSET", range.offset());
        writeDuration(el, "DURATION", range.duration());
        writeLong(el, "STEP", range.step());
        return el;
    }
    public static TimeRange readTimeRangeEl(SS3Database db, Element el) {
        var id = read(el, "ID");
        var name = read(el, "NAME");
        var parent = readRefOpt(el, "PARENT", TIMERANGE_TAG)
            .map(db::getTimeRange);

        var unit = Optional.ofNullable(db.getTimeUnit(readRef(el, TIMEUNIT_TAG)))
            .orElseThrow();

        var offset = parent.map(p -> readDuration(db, el, "OFFSET")).orElse(Duration.ZERO);
        var duration = readDuration(db, el, "DURATION");
        var step = readLong(el, "STEP");

        return TimeRange.ABSOLUTE
            .subrange(id, name, unit, offset, duration, step);
    }

    public static final String TIMEUNIT_LIST_TAG = "TIMEUNITS";
    public static final String TIMEUNIT_TAG = "TIMEUNIT";
    public static Element createTimeUnitEl(Document doc, TimeUnit unit) {
        var el = doc.createElement(TIMEUNIT_TAG);
        write(el, "ID", unit.id());
        write(el, "NAME", unit.name());
        write(el, "SUFFIX", unit.suffix());
        write(el, "TYPE", unit.type().toString());
        writeDouble(el, "BASE", unit.base());
        return el;
    }
    public static TimeUnit readTimeUnitEl(Element el) {
        var id = read(el, "ID");
        var name = read(el, "NAME");
        var suffix = read(el, "SUFFIX");
        var type = TimeUnit.Type.valueOf(read(el, "TYPE"));
        var base = readDouble(el, "BASE");

        return new TimeUnit(id, name, suffix, type, base);
    }




    public static final String VIDEO_LIST_TAG = "VIDEOS";
    public static final String VIDEO_TAG = "VIDEO";
    public static Element writeVideoEl(Document doc, Video video) {
        var el = doc.createElement(VIDEO_TAG);
        write(el, "URI", video.uri().toString());
        writeRef(el, TIMERANGE_TAG, video.range().id());
        writeDuration(el, "DURATION", video.duration());
        writeLong(el, "WIDTH", video.width());
        writeLong(el, "HEIGHT", video.height());
        writeDouble(el, "FRAMERATE", video.framerate());
        return el;
    }
    public static VideoInfo readVideoEl(SS3Database db, Element el) {
        var uri = URI.create(read(el, "URI"));
        var orderStr = read(el, "ORDER");
        var order = Integer.parseInt(orderStr);
        var range = db.getTimeRange(readRef(el, TIMERANGE_TAG));
        var duration = readDuration(db, el, "DURATION");
        var width = (int)readLong(el, "WIDTH");
        var height = (int)readLong(el, "HEIGHT");
        var framerate = readDouble(el, "FRAMERATE");

        return new VideoInfo(uri, order, duration, range, width, height, framerate);
    }

    public static final String AUDIO_LIST_TAG = "AUDIOS";
    public static final String AUDIO_TAG = "AUDIO";
    public static Element writeAudioEl(Document doc, Audio audio) {
        var el = doc.createElement(AUDIO_TAG);
        write(el, "URI", audio.uri().toString());
        writeRef(el, TIMERANGE_TAG, audio.range().id());
        writeDuration(el, "DURATION", audio.duration());
        writeLong(el, "RATE", audio.rate());
        return el;
    }
    public static AudioInfo readAudioEl(SS3Database db, Element el) {
        var uri = URI.create(read(el, "URI"));
        var orderStr = read(el, "ORDER");
        var order = Integer.parseInt(orderStr);
        var range = db.getTimeRange(readRef(el, TIMERANGE_TAG));
        var duration = readDuration(db, el, "DURATION");
        var rate = (int)readLong(el, "RATE");

        return new AudioInfo(uri, order, range, duration, rate);
    }

    public static final String GRAPH_LIST_TAG = "GRAPHS";
    public static final String GRAPH_TAG = "GRAPH";
    public static Element writeGraphEl(Document doc, Graph graph) {
        var el = doc.createElement(GRAPH_TAG);
        write(el, "ID", graph.id());
        write(el, "URI", graph.uri().toString());
        write(el, "NAME", graph.name());
        writeRef(el, TIMERANGE_TAG, graph.range().id());
        return el;
    }
    public static Graph readGraphEl(SS3Database db, Element el) {
        var id = read(el, "ID");
        var name = read(el, "NAME");
        var uri = URI.create(read(el, "URI"));
        var range = db.getTimeRange(readRef(el, TIMERANGE_TAG));

        return new Graph(id, name, uri, range);
    }

//    public static SS3Field readFieldAttributeEl(SS3Database db, Element el) {
//    	
//    	SS3Field field = db.getSS3CodingScheme().getField(fieldID);
//        var id = read(el, "ID");
//        var name = read(el, "NAME");
//        var uri = URI.create(read(el, "URI"));
//
//        return new Graph(id, name, uri, range);
//    }


    /* Coding Schema */


    public static final String PANEL_LIST_TAG = "PANELS";
    public static final String PANEL_TAG = "PANEL";
    public static Panel readPanelEl(Element el) {
        var id = read(el, "ID");
        var label = read(el, "LABEL");
        return new Panel(id, label);
    }

    public static final String VIEW_MODE_LIST_TAG = "VIEW-MODES";
    public static final String VIEW_MODE_TAG = "VIEW-MODE";
    public static ViewMode readViewModeEl(SS3CodingScheme schema, Element el) {
        var id = read(el, "ID");
        var label = read(el, "LABEL");
        var panels = readRefsOrdered(el, PANEL_TAG)
            .map(schema::panel)
            .collect(Collectors.toSet());

        return new ViewMode(id, label, panels);
    }

    public static String getCharacterDataFromElement(Element e) {
        Node child = e.getFirstChild();
        if (child instanceof CharacterData) {
            CharacterData cd = (CharacterData) child;
            return cd.getData();
        }
        return "";
    }
    
    public static void setCharacterDataOnElement(Element e, String str) {
    	Node child = e.getFirstChild();
    	if (child instanceof CharacterData) {
    		child.setNodeValue(str);
    	}
    }
}
