package edu.bu.signstream.media.gstreamer;

import edu.bu.signstream.grepresentation.requestsHandler.SS3Singleton;
import org.freedesktop.gstreamer.Caps;
import org.freedesktop.gstreamer.FlowReturn;
import org.freedesktop.gstreamer.Sample;
import org.freedesktop.gstreamer.elements.AppSink;

import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.util.concurrent.locks.ReentrantLock;

public class GstVideoImage extends JComponent {
    private BufferedImage image;
    private final ReentrantLock lock = new ReentrantLock();

    public GstVideoImage(AppSink sink) {
        sink.setCaps(new Caps("video/x-raw,pixel-aspect-ration=1/1,format=BGRx"));

        sink.set("emit-signals", true);
        sink.connect((AppSink.NEW_SAMPLE) s -> render(s.pullSample()));
        sink.connect((AppSink.NEW_PREROLL) s -> render(s.pullPreroll()));
        sink.connect((AppSink.EOS) elem -> {
            var player = SS3Singleton.getMediaPlayer();
            if (player.time().equals(player.end())) {
                image = null;
                repaint();
            }
        });
    }

    private FlowReturn render(Sample sample) {
        var struct = sample.getCaps().getStructure(0);
        int w = struct.getInteger("width");
        int h = struct.getInteger("height");

        // (Re)allocate BufferedImage if needed
        if (image == null) {
            image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        } else if (image.getWidth() != w || image.getHeight() != h) {
            image.flush();
            image = new BufferedImage(w, h, BufferedImage.TYPE_INT_RGB);
        }

        // Map the sample and render it to the image buffer
        var buf = sample.getBuffer();
        var view = buf.map(false);
        lock.lock();
        try {
            int[] data = ((DataBufferInt)image.getRaster().getDataBuffer()).getData();
            view.asIntBuffer().get(data, 0, w * h);
        } finally {
            lock.unlock();
        }
        buf.unmap();
        sample.dispose();

        repaint();

        return FlowReturn.OK;
    }

    @Override
    protected void paintComponent(Graphics g) {
        var g2d = (Graphics2D)g;

        int realWidth = getWidth();
        int realHeight = getHeight();

        int prefWidth = getPreferredSize().width;
        int prefHeight = getPreferredSize().height;

        double aspect = (double)prefWidth / prefHeight;

        int w = realWidth;
        int h = realHeight;
        int x = 0;
        int y = 0;

        int scaledHeight = (int)Math.round(realWidth / aspect);
        if (scaledHeight < realHeight) {
            y = (realHeight - scaledHeight) / 2;
            h = scaledHeight;
        } else {
            int scaledWidth = (int)Math.round(realHeight * aspect);
            x = (realWidth - scaledWidth) / 2;
            w = scaledWidth;
        }

        g2d.setColor(Color.BLACK);
        g2d.fillRect(0, 0, realWidth, realHeight);

        if (image != null) {
            lock.lock();
            try {
                g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR);
                g2d.drawImage(image, x, y, w, h, null);
            } finally {
                lock.unlock();
            }
        }
    }
}
