package edu.bu.signstream.time;

public class Time implements Comparable<Time> {
    private final TimeRange range;
    private final long value;
    private final TimeUnit unit;

    /**
     * `TimeInfo`s should only be constructed publicly by `TimeRange.time`.
     */
    Time(TimeRange range, long time, TimeUnit unit) {
        this.range = range;
        this.value = time;
        this.unit = unit;
    }

    public TimeRange range() { return range; }
    // JACK TODO: Remove all calls to this, replace with as(TimeUnit.TIMESTAMP)
    // Leaving for now since looking at usages could be useful.
    @Deprecated
	public double doubleValue() {
        return as(TimeUnit.TIMESTAMP, Rounding.NEAREST);
    }
	public long value() { return value; }
	public TimeUnit unit() { return unit; }


    /**
     * Converts the relative time into an integer number of the target unit.
     */
    public long as(TimeUnit unit, Rounding r) {
        return this.unit.convert(value, unit, r);
    }
    /**
     * Converts the time relative to the start of the range into the target unit.
     */
    public double as(TimeUnit unit) {
        return this.unit.convert(value, unit);
    }

    /**
     * Converts the relative time into an integer number of the target unit,
     * and returns a new Time in the same range with the new value and unit.
     */
    public Time convert(TimeUnit unit, Rounding round) {
        return new Time(range, as(unit, round), unit);
    }

    public Time convert(TimeUnit unit) {
        return new Time(range, as(unit, Rounding.NEAREST), unit);
    }

    /**
     * Rebase the TimeInfo onto another TimeRange, preserving the absolute time relative to the absolute range.
     */
    public Time in(TimeRange range, Rounding r) {
        var time = convert(TimeUnit.NANOSECOND, r)
            .add(range().start().duration(), r)
            .subtract(range.start().duration(), r)
            .convert(range.unit(), r);

        return new Time(range, time.value(), range.unit());
    }
    public Time in(TimeRange range) {
        var time = convert(TimeUnit.NANOSECOND)
            .add(range().start().duration())
            .subtract(range.start().duration())
            .convert(range.unit());

        return new Time(range, time.value(), range.unit());
    }

    /**
     * Return this Time in the ABSOLUTE range.
     */
    public Time absolute() {
        return in(TimeRange.ABSOLUTE);
    }

    /**
     * Duration of the Time from the start of the range.
     */
    public Duration duration() {
        return new Duration(value(), unit);
    }

    /**
     * Construct an interval from `this` to another Time.
     */
    public Interval to(Time o) {
        return new Interval(this, o);
    }

    /**
     * Construct an interval from `this` with the specified duration.
     */
    public Interval to(Duration d) { return new Interval(this, this.add(d)); }

    /*
     * All arithmetic functions first convert the value being added to an integer number of the current unit
     * using `Time.as`, then add/subtract the resulting value with the current relative time.
     *
     *  TODO: Remove `Math.round` calls when `value` is migrated to `long`
     */

    public Time add(long offset) {
        return new Time(range, value() + offset, unit);
    }
    public Time add(Duration d, Rounding r) { return new Time(range, value() + d.as(unit, r), unit); }
    public Time add(Duration d) {
        return new Time(range, value() + d.as(unit, Rounding.NEAREST), unit); }

    public Time subtract(long offset) {
        return new Time(range, value() - offset, unit);
    }
    public Time subtract(Duration d, Rounding r) { return new Time(range, value() - d.as(unit, r), unit); }
    public Time subtract(Duration d) {
        return new Time(range, value() - d.as(unit, Rounding.NEAREST), unit); }

    public Time divide(double divisor) {
        return new Time(this.range, Math.round(this.value() / divisor), this.unit);
    }

    public Duration subtract(Time o, Rounding r) {
        return new Duration(value() - o.as(unit, r), unit);
    }
    public Duration subtract(Time o) {
        return new Duration(value() - o.as(unit, Rounding.NEAREST), unit);
    }


    /**
     * Returns the absolute Duration in nanoseconds to another Time.
     */
    public Duration distance(Time o) {
        return new Duration(Math.abs(absolute().value() - o.absolute().value()), TimeUnit.NANOSECOND);
    }

    public Time clamp(Interval i) {
        if (lessThan(i.t0))
            return i.t0.in(range());
        if (greaterThan(i.t1))
            return i.t1.in(range());
        return this;
    }


    /*
     * All comparison functions are done relative to the absolute TimeRange.
     */

    @Override
    public int compareTo(Time o) {
        return Long.compare(absolute().value(), o.absolute().value());
    }

    @Override
    public boolean equals(Object o) { return o instanceof Time && compareTo((Time) o) == 0; }

    public boolean greater(Time o, Bound b) { return b == Bound.INCLUSIVE ? compareTo(o) >= 0 : compareTo(o) > 0; }
    public boolean greaterThan(Time o) { return greater(o, Bound.EXCLUSIVE); }
    public boolean greaterThanOrEq(Time o) { return greater(o, Bound.INCLUSIVE); }

    public boolean less(Time o, Bound b) { return b == Bound.INCLUSIVE ? compareTo(o) <= 0 : compareTo(o) < 0; }
    public boolean lessThan(Time o) { return less(o, Bound.EXCLUSIVE); }
    public boolean lessThanOrEq(Time o) { return less(o, Bound.INCLUSIVE); }

    public Time max(Time o) { return greaterThanOrEq(o) ? this : o; }
    public Time min(Time o) { return lessThanOrEq(o) ? this : o; }


    @Override
    public String toString() {
        return "Time { " + value + " " + unit.suffix() + " " + range.name() + " }";
    }


//    /**
//	 * Returns the value of this TimeInfo in terms of the provided TimeUnit. This does not change the saved unit for this TimeInfo object.
//	 * @param unit the TimeUnit to convert to before returning
//	 * @return
//	 */
//	public double getUnitValueAs(TimeUnit unit) {
//        return unit().convert(value, unit);
//
//		double packetDuration = SS3Singleton.getSignStreamApplication().getSignStreamSegmentPanel().getMultipleMovieController().getPacketDuration();
//		double ret = value();
//
//		if (unit().equals(TimeUnit.VIDEO_FRAME_30_FPS)) {
//			ret /= packetDuration;
//		}
//
//		ret = unit().convertTo(ret, unit);
//
//		if (unit.equals(TimeUnit.VIDEO_FRAME_30_FPS)) {
//			ret *= packetDuration;
//			ret = closestNumber(ret, packetDuration);
//		}
//
//		return ret;
//	}

//	/**
//	 * Useful for getting the frame boundaries for a given 'frame duration'
//	 * @param a
//	 * @param b
//	 * @return
//	 */
//	private static double closestNumber(double a, double b) {
//		double c1 = a - (a % b);
//		double c2 = (a + b) - (a % b);
//
//		if (a - c1 > c2 - a) {
//			return c2;
//		} else {
//			return c1;
//		}
//	}

//	/**
//	 * Sets the unit to the provided TimeUnit, and sets the appropriate converted value according to the provided TimeUnit
//	 * @param unit the TimeUnit to convert to and save as
//	 */
//	public void setUnitAndValueAs(TimeUnit unit) {
//        this.setUnitValue(getUnitValueAs(unit));
//		this.setTimeUnit(unit);
//	}

//	/**
//	 * Changes the unit by which this TimeInfo is measured.
//	 * The raw value stays the same.
//	 * @param unit the new TimeUnit to use.
//	 */
//	public void setTimeUnit(TimeUnit unit) {
//		this.unit = unit;
//		updateDependentUnitValues();
//	}
	
//	/**
//	 * Changes the unit by which this TimeInfo is measured. If provided string is either null or invalid, deafult 30FPS type will be chosen
//	 * The raw value statys the same.
//	 * @param str the string value of one of the enum types for TimeUnit
//	 */
//	public void setTimeUnit(String str) {
//		TimeUnit unit;
//		try {
//			unit = TimeUnit.valueOf(str);
//		} catch (IllegalArgumentException e) {
//			System.out.println("Error from TimeInfo.setTimeUnit: " + e);
//			unit = TimeUnit.VIDEO_FRAME_30_FPS;		// default setting for time unit is 30 fps
//		}
//		setTimeUnit(unit);
//	}

	/****************/


	// JACK TODO: Just update dependent Times when making any changes to entities instead.
	public void addDependentTimeInfo(Time timeInfo) {}
//		dependentTimeInfos.add(timeInfo);
//	}
//	public void removeTimeInfo(TimeInfo timeInfo) {
//		dependentTimeInfos.remove(timeInfo);
//	}
//	private void updateDependentUnitValues() {
//		for(TimeInfo dTimeInfo : dependentTimeInfos) {
//			dTimeInfo.setFrameNumber(frameNumber);
//			dTimeInfo.setUnitValue(value);
//			dTimeInfo.setTimeUnit(unit);
//		}
//	}


	// Test functionality
	public static void main(String[] args) {
//	    TimeRange global = TimeRange.ABSOLUTE;
//	    TimeRange r = global.subrange(TimeUnit.MILLISECOND, Duration.ms(500));

//	    var t = r.time(1000);
//		System.out.println("orig unit: " + t.unit());
//		System.out.println("unit value: " + t.value());
//        System.out.println("in absolute: " + t.in(TimeRange.ABSOLUTE).value());
//		System.out.println("value as: " + t.getUnitValueAs(TimeUnit.MICROSECOND));
		
//		System.out.println("long max value: " + Long.MAX_VALUE);
//		System.out.println("double max value: " + Double.MAX_VALUE);
//		System.out.println("double max exponent: " + Double.MAX_EXPONENT);
//		System.out.println("long max value casted to double: " + (double) Long.MAX_VALUE);		// widening conversion
//		// https://docs.oracle.com/javase/specs/jls/se7/html/jls-5.html
		
			
		
		//SS3Database s;
		//s.setStartTimeInfo(new Time(1, TimeUnit.SECOND));
	}

    // JACK TODO: Remove this
    // not sure how this is used - I see usages from SS2, not relevant in SS3, though
    private int frameNumber = -1;
    @Deprecated public int getFrameNumber() {
        return frameNumber;
    }
    @Deprecated public void setFrameNumber(int i) { frameNumber = i; }
}
