package edu.bu.signstream.ui.panels.search;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Set;

import edu.bu.signstream.common.util.Util;
import edu.bu.signstream.common.util.vo.Constants;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3Field;
import edu.bu.signstream.common.util.vo.ss3.codingScheme.SS3FieldValue;
import edu.bu.signstream.common.util.vo.ss3.db.SS3Database;
import edu.bu.signstream.common.util.vo.ss3.db.SS3Entity;
import edu.bu.signstream.common.util.vo.ss3.db.SS3GlossWindowItem;
import edu.bu.signstream.common.util.vo.ss3.db.SS3GlossWindowSegment;
import edu.bu.signstream.common.util.vo.ss3.db.SS3Item;
import edu.bu.signstream.common.util.vo.ss3.db.SS3StatementField;
import edu.bu.signstream.common.util.vo.ss3.db.SS3TemporalPartition;

public class SearchEngine {
	
	private SearchTreeNode rootSearchTreeNode = null;
	private SS3Database database = null;
	private ArrayList<SS3GlossWindowItem> fullItemList; 
	private HashMap<String, SS3GlossWindowItem> itemIDMap = null;
	
	public SearchEngine(SS3Database database) {
		this.database = database;
		
		// on construction, create the list containing all items in the DB
		if (database != null) {
			fullItemList = new ArrayList<SS3GlossWindowItem>();
			itemIDMap = new HashMap<String, SS3GlossWindowItem>();

			ArrayList<SS3TemporalPartition> temporalPartitions = database.getSS3GlossWindow().getTemporalPartitions();
			
			for (SS3TemporalPartition temporalPartition: temporalPartitions) {
				ArrayList<SS3GlossWindowSegment> segmentTiers = temporalPartition.getTemporalPartitionSegments();
				for (SS3GlossWindowSegment segmentTier: segmentTiers) {
					ArrayList<SS3GlossWindowItem> items = segmentTier.getSS3GlossWindowItems();
					fullItemList.addAll(items);
					
					for (SS3GlossWindowItem item: items) {
						itemIDMap.put(item.getId(), item);
					}
				}
			}
		}
	}
	
	public ArrayList<SS3GlossWindowItem> search(SearchTreeNode rootSearchTreeNode) {
		this.rootSearchTreeNode = rootSearchTreeNode;
		
		return searchRecursively(rootSearchTreeNode);
	}
	
	private ArrayList<SS3GlossWindowItem> searchRecursively(SearchTreeNode searchTreeNode) {
		
		ArrayList<SS3GlossWindowItem> items = new ArrayList<SS3GlossWindowItem>();
		
		SearchTreeNode firstChildNode = null, secondChildNode = null;
		if (searchTreeNode.getChildCount() == 2) {
			firstChildNode = (SearchTreeNode) searchTreeNode.getFirstChild(); 
			secondChildNode = (SearchTreeNode) searchTreeNode.getLastChild(); 
		}

		SearchOperator searchOperator = searchTreeNode.getSearchOperator();
		switch (searchOperator) {
			case AND:
				items = searchAnd(searchTreeNode);
				break;
			case OR:
				items = searchOr(searchTreeNode);
				break;
			case WITH:
			case BEFORE:
			case AFTER:
				items = searchBinaryGroup(searchOperator, firstChildNode, secondChildNode);
				break;
			default:
				break;
		}
		
		// perform the subtraction from the DB item list to accomplish negation of this group
		if (searchTreeNode.isNegateGroup()) {
			ArrayList<SS3GlossWindowItem> allOtherItems = new ArrayList<SS3GlossWindowItem>(fullItemList);
			allOtherItems.removeAll(items);
			return allOtherItems;
		}
		
		return items;
	}

	// searchBinaryGroup is a heavily modified searchRule, but it still retains most of its functionality
	// if there are changes in searchRule, be sure to make the appropriate corresponding change in searchBinaryGroup
	private ArrayList<SS3GlossWindowItem> searchRule(SearchExpressionData data) {
		ArrayList<SS3GlossWindowItem> results = new ArrayList<SS3GlossWindowItem>();
		ArrayList<SS3GlossWindowItem> allOtherItems = new ArrayList<SS3GlossWindowItem>(fullItemList);
		ArrayList<SS3GlossWindowItem> allOtherItems2 = new ArrayList<SS3GlossWindowItem>(fullItemList);
		ArrayList<SS3GlossWindowItem> allOtherItems3 = new ArrayList<SS3GlossWindowItem>(fullItemList);

		boolean validData = data.isValidData();
		boolean negateField = data.isNegateField();
		boolean negateFieldAndValue = data.isNegateFieldAndValue();
		boolean negateParticipant = data.isNegateParticipant();
		boolean negateRule = data.isNegateRule();
		boolean negateValue = data.isNegateValue();		
		boolean exactWord = data.isExactWord();
		boolean matchCase = data.isMatchCase();		
		//		String participantID = data.getParticipantID();
		boolean anyParticipant = data.isAnyParticipant();
		String participantName = data.getParticipantName();
		boolean anyField = data.isAnyField();
		String fieldID = data.getFieldID();
		String fieldName = data.getFieldName();
		//		String valueID = data.getValueID();
		boolean anyValue = data.isAnyValue();
		String valueName = data.getValueName();
		String valueText = data.getValueText().trim();
		//		String frameSearchBy = data.getFrameSearchBy();

		if ( !validData) {
			return results;
		}

		ArrayList<SS3TemporalPartition> temporalPartitions = database.getSS3GlossWindow().getTemporalPartitions();
		ArrayList<SS3GlossWindowSegment> segmentTiers;
		ArrayList<SS3StatementField> fields;
		ArrayList<SS3StatementField> filteredFields;
		ArrayList<SS3Entity> entities;
		ArrayList<SS3Item> entityItems;

		for (SS3TemporalPartition temporalPartition: temporalPartitions) {
			segmentTiers = temporalPartition.getTemporalPartitionSegments();

			for (SS3GlossWindowSegment segmentTier: segmentTiers) {
				// look at only the segment tiers matching the participant selected				
				String segmentParticipant = segmentTier.getParticipantLabel();
				boolean segmentHasParticipant = segmentParticipant != null;				
				boolean segmentHasThisParticipant = false;
				if (participantName != null) {
					segmentHasThisParticipant = participantName.equals(segmentParticipant);
				}
				
				if (negateParticipant) {
					if (anyParticipant) {		// "if not any participant"
						continue;
					} else {
						if (segmentHasParticipant && segmentHasThisParticipant) {	// "if not this participant"
							continue;
						}
					}
				} else {
					if (!anyParticipant) {
						if (segmentHasParticipant && !segmentHasThisParticipant) {	// "if this participant does not match this segment's participant"
							continue;
						}
					}
				}

				fields = new ArrayList<SS3StatementField>( segmentTier.getStatementFields());	// make a copy of the list, due to the potential removeAll below
				filteredFields = new ArrayList<SS3StatementField>();

				if (!fields.isEmpty()) {
					if (SearchRule.ANY_GLOSS_FIELD_NAME.equals(fieldName)) {
						filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GLOSS_FIELD_ID));
						filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GLOSS_FIELD_ID));
					} else if (SearchRule.DOM_GLOSS_FIELD_NAME.equals(fieldName)) {
						filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GLOSS_FIELD_ID));
					} else if (SearchRule.NDOM_GLOSS_FIELD_NAME.equals(fieldName)) {
						filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GLOSS_FIELD_ID));
                    } else if (SearchRule.ANY_GES_FIELD_NAME.equals(fieldName)) {
                        filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GESTURE_FIELD_ID));
                        filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GESTURE_FIELD_ID));
                    } else if (SearchRule.DOM_GES_FIELD_NAME.equals(fieldName)) {
                        filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GESTURE_FIELD_ID));
                    } else if (SearchRule.NDOM_GES_FIELD_NAME.equals(fieldName)) {
                        filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GESTURE_FIELD_ID));
					} else if (fieldID != null) {
						for (SS3StatementField field : fields) {
							if (field.getId().equals(fieldID)) {
								filteredFields.add(field);
							}
						}
					} else {
						filteredFields = fields;
					}
				}

				if (fieldID != null && negateField) {
					fields.removeAll(filteredFields);
					filteredFields = fields;
				}

				boolean matchFound = negateValue ? true : false;

				for (SS3StatementField field: filteredFields) {
					SS3Field dbField = database.getSS3CodingScheme().getField(field.getId());
					
					matchFound = negateValue ? true : false;
					
					entities = field.getSS3Entities();
					if (entities.isEmpty()) {
						matchFound = false;
						continue;	// if this field is empty, don't bother looking at it, even with negation
					}

					for (SS3Entity entity: entities) {
						matchFound = negateValue ? true : false;

						entityItems = entity.getItems();

						for (SS3Item entityItem: entityItems) {
							if (entityItem.isTextOnly())
								continue;
							
							if (anyValue) {
								matchFound = negateValue ? false : true;
							} else {
								String itemText = entityItem.getText().trim();
								if (entity.getUniqueNumber() != null)
								    itemText += "-" + entity.getUniqueNumber();

								if ((valueText == null || valueText.isEmpty()) && 
										dbField.getType().equals(Constants.TYPE_NON_MANUAL)) {
									String valueID = null;
									for (SS3FieldValue value: dbField.getValuesAsList()) {
										if (value.getName().equals(valueName)) {
											valueID = value.getId();
											break;
										}
									}
									if (valueID != null && valueID.equals(entityItem.getId())) {
										matchFound = negateValue ? false : true;
									}
								} else {
									String ruleText = null;

									// compare using the text of either what is typed in the textbox or what is selected in the dropdown
									if (valueText != null && !valueText.isEmpty())
										ruleText = valueText;
									else {
										ruleText = valueName;
										exactWord = true;    // setting this flag true so that we match valueName exactly
										// e.g. don't mistake "-right" for "right" by checking for the exact word
									}

									if (!matchCase) {
										ruleText = ruleText.toLowerCase();
										itemText = itemText.toLowerCase();
									}

									if (exactWord) {
										if (itemText.equals(ruleText)) {
											matchFound = negateValue ? false : true;
										}
									} else {
										if (itemText.contains(ruleText)) {
											matchFound = negateValue ? false : true;
										}
									}
								}
							}
						}	// entity items

						ArrayList<String> parentEntityIDs = entity.getParentEntityIDs();
						ArrayList<SS3GlossWindowItem> items = new ArrayList<SS3GlossWindowItem>();
						for (String parentEntityID: parentEntityIDs) {
							SS3GlossWindowItem item = itemIDMap.get(parentEntityID);
							items.add(item);
						}

						for (SS3GlossWindowItem item: items) {
							if (matchFound && !negateFieldAndValue) {
								if ( item != null && !results.contains(item)) {
									results.add(item);
								}
							} else if ( !matchFound && negateFieldAndValue) {
								if ( item != null && !results.contains(item)) {
									results.add(item);
								}
							}
						}

					}	// entities

				}	// fields

			}
		}

		// in the case of negateRule (meaning literally negate what this rule says to do):
		// return all of other items in this database which are not the ones we matched and put in the results list
		if (negateRule) {
			allOtherItems.removeAll(results);
			return allOtherItems;
		}
		
		// to preserve ordering...
		allOtherItems2.removeAll(results);
		allOtherItems3.removeAll(allOtherItems2);
		return allOtherItems3;
	}
	
	private ArrayList<SS3GlossWindowItem> searchAnd(SearchTreeNode node) {
		ArrayList<ArrayList<SS3GlossWindowItem>> lists = new ArrayList<ArrayList<SS3GlossWindowItem>>();
		
		for (SearchTreeNode child: node.getChildrenAsList()) {
			if (child.isRuleNode()) {
				SearchExpressionData data = child.getData();
				lists.add(searchRule(data));
			} else if (child.isGroupNode()) {
				lists.add(searchRecursively(child));
			}
		}
		
		return intersection(lists);
	}
	
	private ArrayList<SS3GlossWindowItem> searchOr(SearchTreeNode node) {
		ArrayList<ArrayList<SS3GlossWindowItem>> lists = new ArrayList<ArrayList<SS3GlossWindowItem>>();
		
		for (SearchTreeNode child: node.getChildrenAsList()) {
			if (child.isRuleNode()) {
				SearchExpressionData data = child.getData();
				lists.add(searchRule(data));
			} else if (child.isGroupNode()) {
				lists.add(searchRecursively(child));
			}
		}
		
		return union(lists);
	}
	
	// searchBinaryGroup is a heavily modified searchRule, but it still retains most of its functionality
	// if there are changes in searchRule, be sure to make the appropriate corresponding change in searchBinaryGroup
	private ArrayList<SS3GlossWindowItem> searchBinaryGroup(SearchOperator operator, SearchTreeNode node, SearchTreeNode node2) {		
		ArrayList<SS3GlossWindowItem> results = new ArrayList<SS3GlossWindowItem>();
		ArrayList<SS3GlossWindowItem> allOtherItems2 = new ArrayList<SS3GlossWindowItem>(fullItemList);
		ArrayList<SS3GlossWindowItem> allOtherItems3 = new ArrayList<SS3GlossWindowItem>(fullItemList);

		SearchExpressionData data = node.getData();
		boolean validData = data.isValidData();
		boolean negateField = data.isNegateField();
//		boolean negateFieldAndValue = data.isNegateFieldAndValue();
		boolean negateParticipant = data.isNegateParticipant();
//		boolean negateRule = data.isNegateRule();
		boolean negateValue = data.isNegateValue();		
		boolean exactWord = data.isExactWord();
		boolean matchCase = data.isMatchCase();		
//		String participantID = data.getParticipantID();
		boolean anyParticipant = data.isAnyParticipant();
		String participantName = data.getParticipantName();
		boolean anyField = data.isAnyField();
		String fieldID = data.getFieldID();
		String fieldName = data.getFieldName();
//		String valueID = data.getValueID();
		boolean anyValue = data.isAnyValue();
		String valueName = data.getValueName();
		String valueText = data.getValueText();
		String frameSearchBy = data.getFrameSearchBy();
		
		if ( !validData) {
			return results;
		}
		
		ArrayList<SS3TemporalPartition> temporalPartitions = database.getSS3GlossWindow().getTemporalPartitions();
		ArrayList<SS3GlossWindowSegment> segmentTiers;
		ArrayList<SS3StatementField> fields;
		ArrayList<SS3StatementField> filteredFields;
		ArrayList<SS3Entity> entities;
		ArrayList<SS3Item> entityItems;
		
		for (SS3TemporalPartition temporalPartition: temporalPartitions) {
			 segmentTiers = temporalPartition.getTemporalPartitionSegments();
			 
			 for (SS3GlossWindowSegment segmentTier: segmentTiers) {
					// look at only the segment tiers matching the participant selected				
					String segmentParticipant = segmentTier.getParticipantLabel();
					boolean segmentHasParticipant = segmentParticipant != null;				
					boolean segmentHasThisParticipant = false;
					if (participantName != null) {
						segmentHasThisParticipant = participantName.equals(segmentParticipant);
					}
					
					if (negateParticipant) {
						if (anyParticipant) {		// "if not any participant"
							continue;
						} else {
							if (segmentHasParticipant && segmentHasThisParticipant) {	// "if not this participant"
								continue;
							}
						}
					} else {
						if (!anyParticipant) {
							if (segmentHasParticipant && !segmentHasThisParticipant) {	// "if this participant does not match this segment's participant"
								continue;
							}
						}
					}
				 
					 boolean comparisonSuccess = false;
					 
					 fields = segmentTier.getStatementFields();

					 filteredFields = new ArrayList<SS3StatementField>();

					 if (SearchRule.ANY_GLOSS_FIELD_NAME.equals(fieldName)) {
						 filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GLOSS_FIELD_ID));
						 filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GLOSS_FIELD_ID));
					 } else if (SearchRule.DOM_GLOSS_FIELD_NAME.equals(fieldName)) {
						 filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GLOSS_FIELD_ID));
					 } else if (SearchRule.NDOM_GLOSS_FIELD_NAME.equals(fieldName)) {
						 filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GLOSS_FIELD_ID));
                     } else if (SearchRule.ANY_GES_FIELD_NAME.equals(fieldName)) {
                         filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GESTURE_FIELD_ID));
                         filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GESTURE_FIELD_ID));
                     } else if (SearchRule.DOM_GES_FIELD_NAME.equals(fieldName)) {
                         filteredFields.add(segmentTier.getSS3StatementField(Util.DOM_GESTURE_FIELD_ID));
                     } else if (SearchRule.NDOM_GES_FIELD_NAME.equals(fieldName)) {
                         filteredFields.add(segmentTier.getSS3StatementField(Util.NDOM_GESTURE_FIELD_ID));
					 } else if (fieldID != null) {
						 for (SS3StatementField field: fields) {
							 if (field.getId().equals(fieldID)) {
								 filteredFields.add(field);
							 }
						 }
					 } else {
						 filteredFields = fields;
					 }

					 if (fieldID != null && negateField) {
						 fields.removeAll(filteredFields);
						 filteredFields = fields;
					 }
					 
					 boolean matchFound = negateValue ? true : false;
					 
                     SS3Entity matchingEntity = null;
					 for (SS3StatementField field: filteredFields) {
						 SS3Field dbField = database.getSS3CodingScheme().getField(field.getId());
						 
						 entities = field.getSS3Entities();
						 if (entities.isEmpty()) {
							 matchFound = false;
							 continue;	// if this field is empty, don't bother looking at it, even with negation
						 }
						 
						 for (SS3Entity entity: entities) {
							 // use the start/end times from the SS3Entity objs since the SS3Item doesn't have time set properly
							 double itemStartTime = entity.getStartTime();
							 double itemEndTime = entity.getEndTime();
							 entityItems = entity.getItems();
							 
							 for (SS3Item entityItem: entityItems) {
								 if (entityItem.isTextOnly())
									 continue;
								 
							     if (anyValue) {
							         matchFound = negateField ? false : true;
                                 } else {
                                     String itemText = entityItem.getText();
                                     if (entity.getUniqueNumber() != null)
                                         itemText += "-" + entity.getUniqueNumber();
                                     
                                     if ((valueText == null || valueText.isEmpty()) &&
                                    		 dbField.getType().equals(Constants.TYPE_NON_MANUAL)) {
                                    	 String valueID = null;
                                    	 for (SS3FieldValue value: dbField.getValuesAsList()) {
                                    		 if (value.getName().equals(valueName)) {
                                    			 valueID = value.getId();
                                    			 break;
                                    		 }
                                    	 }
                                    	 if (valueID.equals(entityItem.getId())) {
                                             matchFound = negateValue ? false : true;
                                    	 }
                                     } else {
                                         String ruleText = null;

                                         // compare using the text of either what is typed in the textbox or what is selected in the dropdown
                                         if (valueText != null && !valueText.isEmpty())
                                             ruleText = valueText;
                                         else {
                                             ruleText = valueName;
                                             exactWord = true;    // setting this flag true so that we match valueName exactly
                                             // e.g. don't mistake "-right" for "right" by checking for the exact word
                                         }

                                         if (!matchCase) {
                                             ruleText = ruleText.toLowerCase();
                                             itemText = itemText.toLowerCase();
                                         }

                                         if (exactWord) {
                                             if (itemText.equals(ruleText)) {
                                                 matchFound = negateValue ? false : true;
                                             }
                                         } else {
                                             if (itemText.contains(ruleText)) {
                                                 matchFound = negateValue ? false : true;
                                             }
                                         }
                                     }
                                 }
								 
								 comparisonSuccess = false;

								 // if we find a match in 'node' for its rule criteria, then loop again through 'node2' for its own rule criteria
								 if (matchFound) {
									SearchExpressionData data2 = node2.getData();
									boolean validData2 = data2.isValidData();
									boolean negateField2 = data2.isNegateField();
									boolean anyField2 = data2.isAnyField();
//									boolean negateFieldAndValue2 = data2.isNegateFieldAndValue();
//									boolean negateParticipant2 = data2.isNegateParticipant();
//									boolean negateRule2 = data2.isNegateRule();
									boolean negateValue2 = data2.isNegateValue();
									boolean anyValue2 = data2.isAnyValue();
									boolean exactWord2 = data2.isExactWord();
									boolean matchCase2 = data2.isMatchCase();		
//										String participantID2 = data2.getParticipantID();
//									String participantName2 = data2.getParticipantName();
									String fieldID2 = data2.getFieldID();
										String fieldName2 = data2.getFieldName();
//										String valueID2 = data2.getValueID();
									String valueName2 = data2.getValueName();
									String valueText2 = data2.getValueText();
									String frameSearchBy2 = data2.getFrameSearchBy();
									
									if ( !validData2) {
										return results;
									}
									
									ArrayList<SS3StatementField> fields2;
									ArrayList<SS3StatementField> filteredFields2;
									ArrayList<SS3Entity> entities2;
									ArrayList<SS3Item> entityItems2;
									 
									 fields2 = segmentTier.getStatementFields();

									 filteredFields2 = new ArrayList<SS3StatementField>();

									 if (SearchRule.ANY_GLOSS_FIELD_NAME.equals(fieldName2)) {
										 filteredFields2.add(segmentTier.getSS3StatementField(Util.DOM_GLOSS_FIELD_ID));
										 filteredFields2.add(segmentTier.getSS3StatementField(Util.NDOM_GLOSS_FIELD_ID));
									 } else if (SearchRule.DOM_GLOSS_FIELD_NAME.equals(fieldName2)) {
										 filteredFields2.add(segmentTier.getSS3StatementField(Util.DOM_GLOSS_FIELD_ID));
									 } else if (SearchRule.NDOM_GLOSS_FIELD_NAME.equals(fieldName2)) {
										 filteredFields2.add(segmentTier.getSS3StatementField(Util.NDOM_GLOSS_FIELD_ID));
                                     } else if (SearchRule.ANY_GES_FIELD_NAME.equals(fieldName2)) {
                                         filteredFields2.add(segmentTier.getSS3StatementField(Util.DOM_GESTURE_FIELD_ID));
                                         filteredFields2.add(segmentTier.getSS3StatementField(Util.NDOM_GESTURE_FIELD_ID));
                                     } else if (SearchRule.DOM_GES_FIELD_NAME.equals(fieldName2)) {
                                         filteredFields2.add(segmentTier.getSS3StatementField(Util.DOM_GESTURE_FIELD_ID));
                                     } else if (SearchRule.NDOM_GES_FIELD_NAME.equals(fieldName2)) {
                                         filteredFields2.add(segmentTier.getSS3StatementField(Util.NDOM_GESTURE_FIELD_ID));
									 } else if (fieldID2 != null) {
										 for (SS3StatementField field2: fields2) {
											 if (field2.getId().equals(fieldID2)) {
												 filteredFields2.add(field2);
											 }
										 }
									 } else {
										 filteredFields2 = fields2;
									 }

									 if (fieldID2 != null && negateField2) {
										 fields2.removeAll(filteredFields2);
										 filteredFields2 = fields;
									 }
									 
									 boolean matchFound2 = negateValue2 ? true : false;
									 
									 for (SS3StatementField field2: filteredFields2) {
										 entities2 = field2.getSS3Entities();
										 if (entities2.isEmpty()) {
											 matchFound2 = false;
											 continue;	// if this field is empty, don't bother looking at it, even with negation
										 }
										 
										 for (SS3Entity entity2: entities2) {
											 // use the start/end times from the SS3Entity objs since the SS3Item doesn't have time set properly
											 double itemStartTime2 = entity2.getStartTime();
											 double itemEndTime2 = entity2.getEndTime();
											 entityItems2 = entity2.getItems();
											 
											 for (SS3Item entityItem2: entityItems2) {
											     if (anyValue2) {
											         matchFound2 = negateValue2 ? false : true;
                                                 } else {
                                                     String itemText2 = entityItem2.getText();
                                                     if (entity2.getUniqueNumber() != null)
                                                         itemText2 += "-" + entity2.getUniqueNumber();

                                                     String ruleText2 = null;

                                                     // compare using the text of either what is typed in the textbox or what is selected in the dropdown
                                                     if (valueText2 != null && !valueText2.isEmpty())
                                                         ruleText2 = valueText2;
                                                     else {
                                                         ruleText2 = valueName2;
                                                         exactWord2 = true;    // setting this flag true so that we match valueName2 exactly
                                                         // e.g. don't mistake "-right" for "right" by checking for the exact word
                                                     }

                                                     if (!matchCase2) {
                                                         ruleText2 = ruleText2.toLowerCase();
                                                         itemText2 = itemText2.toLowerCase();
                                                     }

                                                     if (exactWord2) {
                                                         if (itemText2.equals(ruleText2)) {
                                                             matchFound2 = negateValue2 ? false : true;
                                                         }
                                                     } else {
                                                         if (itemText2.contains(ruleText2)) {
                                                             matchFound2 = negateValue2 ? false : true;
                                                         }
                                                     }
                                                 }
												 
												 if (matchFound && matchFound2) {
					                                 matchingEntity = entity;

													 comparisonSuccess = binaryOperatorComparator(operator, 
															 frameSearchBy, itemStartTime, itemEndTime, 
															 frameSearchBy2, itemStartTime2, itemEndTime2);
													 
													 // if they do time align, then break out of all the loops and add to results
													 if (comparisonSuccess)
														 break;
												 }
												 
											 }		// entity items 2
											 
											 if (comparisonSuccess)
												 break;
											 else
												 matchFound2 = negateValue2 ? true : false;
										 }
										 
										 if (comparisonSuccess)
											 break;
										 else
											 matchFound2 = negateValue2 ? true : false;
									 }
									 
									 if (comparisonSuccess)
										 break;
									 else
										 comparisonSuccess = false;
								 }
							 }		// entity items
							 
							 if (comparisonSuccess)
								 break;
						 }		// entities
						 
						 if (comparisonSuccess)
							 break;
					 }		// fields

					 if (matchingEntity != null) {
						 ArrayList<String> parentEntityIDs = matchingEntity.getParentEntityIDs();
						 ArrayList<SS3GlossWindowItem> items = new ArrayList<SS3GlossWindowItem>();
						 for (String parentEntityID: parentEntityIDs) {
							 SS3GlossWindowItem item = itemIDMap.get(parentEntityID);
							 items.add(item);
						 }
	
						 for (SS3GlossWindowItem item: items) {
							 if (comparisonSuccess) {
								 if ( !results.contains(item)) {
									 results.add(item);
								 }
							 } 
						 }
					 }
			 }
		}
		
		// to preserve ordering...
		allOtherItems2.removeAll(results);
		allOtherItems3.removeAll(allOtherItems2);
		return allOtherItems3;
	}
	
	private ArrayList<SS3GlossWindowItem> intersection(ArrayList<ArrayList<SS3GlossWindowItem>> lists) {
		ArrayList<SS3GlossWindowItem> results = new ArrayList<SS3GlossWindowItem>();

		int listsSize = lists.size();
		
		if (listsSize == 0) {
			return results;
		} else if (listsSize == 1) {
			results = lists.get(0);
		} else {
			results = lists.get(0);
			
			for (int i = 1; i < listsSize; i++) {
				ArrayList<SS3GlossWindowItem> list = lists.get(i);
				results.retainAll(list);		// keep only items in results which are also in list
			}
		}		
		
		return results;
	}
	
	private ArrayList<SS3GlossWindowItem> union(ArrayList<ArrayList<SS3GlossWindowItem>> lists) {
		Set<SS3GlossWindowItem> setResults = new LinkedHashSet<SS3GlossWindowItem>();
		ArrayList<SS3GlossWindowItem> results = new ArrayList<SS3GlossWindowItem>();

		int listsSize = lists.size();
		
		if (listsSize == 0) {
			return results;
		} else if (listsSize == 1) {
			results = lists.get(0);
		} else {
			for (ArrayList<SS3GlossWindowItem> list: lists) {
				setResults.addAll(list);
			}
			
			results = new ArrayList<SS3GlossWindowItem>(setResults);
		}		
		
		return results;
	}
	
	private boolean binaryOperatorComparator(SearchOperator operator, 
						String searchBy1, double startTime1, double endTime1,
						String searchBy2, double startTime2, double endTime2) {
		
		if (startTime1 < 0 || endTime1 < 0 || startTime1 < 0 || endTime2 < 0)
			return false;
		
		switch (operator) {
		case WITH:
			
			if (startTime1 > endTime2)
				return false;			
			
			if (searchBy1.equals("ANY")) {
				if (searchBy2.equals("ANY")) {
					if (isBetweenInclusive(startTime1, startTime2, endTime2)
						|| isBetweenInclusive(startTime2, startTime1, endTime1)) {
						return true;
					}
				} else if (searchBy2.equals("START")) {
					return isBetweenInclusive(startTime2, startTime1, endTime1);
				} else if (searchBy2.equals("END")) {
					return isBetweenInclusive(endTime2, startTime1, endTime1);
				}
			} else if (searchBy1.equals("START")) {
				if (searchBy2.equals("ANY")) {
					return isBetweenInclusive(startTime1, startTime2, endTime2);
				} else if (searchBy2.equals("START")) {
					return startTime1 == startTime2;
				} else if (searchBy2.equals("END")) {
					return startTime1 == endTime2;
				}				
			} else if (searchBy1.equals("END")) {
				if (searchBy2.equals("ANY")) {
					return isBetweenInclusive(endTime1, startTime2, endTime2);
				} else if (searchBy2.equals("START")) {
					return endTime1 == startTime2;
				} else if (searchBy2.equals("END")) {
					return endTime1 == endTime2;
				}
			}
			
			break;
		case BEFORE:			

			if (searchBy1.equals("ANY")) {
				if (searchBy2.equals("ANY")) {
					return ! (startTime1 > endTime2);
				} else if (searchBy2.equals("START")) {
					return (startTime1 < startTime2);
				} else if (searchBy2.equals("END")) {
					return (startTime1 < endTime2);
				}
			} else if (searchBy1.equals("START")) {
				if (searchBy2.equals("ANY")) {
					return (startTime1 < endTime2);
				} else if (searchBy2.equals("START")) {
					return (startTime1 < startTime2);
				} else if (searchBy2.equals("END")) {
					return (startTime1 < endTime2);
				}
			} else if (searchBy1.equals("END")) {
				if (searchBy2.equals("ANY")) {
					return (endTime1 < endTime2);
				} else if (searchBy2.equals("START")) {
					return (endTime1 < startTime2);
				} else if (searchBy2.equals("END")) {
					return (endTime1 < endTime2);
				}
			}
			
			break;
		case AFTER:
			
			if (searchBy1.equals("ANY")) {
				if (searchBy2.equals("ANY")) {
					return ! (startTime2 > endTime1);
				} else if (searchBy2.equals("START")) {
					return ! (startTime2 > endTime1);
				} else if (searchBy2.equals("END")) {
					return (endTime1 > endTime2);
				}
			} else if (searchBy1.equals("START")) {
				if (searchBy2.equals("ANY")) {
					return (startTime1 > startTime2);
				} else if (searchBy2.equals("START")) {
					return (startTime1 > startTime2);
				} else if (searchBy2.equals("END")) {
					return (startTime1 > endTime2);
				}
			} else if (searchBy1.equals("END")) {
				if (searchBy2.equals("ANY")) {
					return ! (endTime1 < startTime2);
				} else if (searchBy2.equals("START")) {
					return ! (endTime1 < startTime2);
				} else if (searchBy2.equals("END")) {
					return (endTime1 > endTime2);
				}
			}
			
			break;
		default:
			return false;
		}
		
		return false;
	}
	
	private boolean isBetweenInclusive(double frameCheck, double frame1, double frame2) {
		boolean ret = (frameCheck >= frame1 && frameCheck <= frame2);
		return ret;
	}

}
